/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file alsastream.cc
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// std c
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// this header
#include "alsastream.hh"

// PIAVE
#include "libpiave/frame.hh"
#include "libpiave/property.hh"

using namespace PIAVE;

PIAVE_PLUGIN( AlsaPlugin );


/*
 *    wrapper plugin
 */
AlsaPlugin::AlsaPlugin() :
  Plugin( "ALSAStream", PLUGIN_OUTSTREAM )
{
    PropertyNode * caps = new PropertyNode( "outstream" );
    caps->addAttr( "name", getName() );

    PropertyNode * monitor = caps->addDaughter( "monitor" );
    PropertyNode * audio = monitor->addDaughter( "audio" );
    audio->addAttr("type","alsa");

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Simple ALSA layer for audio output.\n" );

    setCapabilities( caps );

    /* this interface is prioritized over e.g. oss */
    setPriority( 10 );

}


PluginBase * 
AlsaPlugin::newElement( const PropertyNode * p )
{
    return new OutALSAStream();
}






/*
 *  \class OutALSAstream
 *
 *
 */
OutALSAStream::OutALSAStream() : 
  OutAudioStreamIFace(),
  _playbackHandle( 0 ),
  _deviceName( "plughw:0,0" ),
  _nChannels( -1 ),
  _frequency( -1 ),
  _warnCount( 20 )
{
}

OutALSAStream::~OutALSAStream()
{
    snd_pcm_close( _playbackHandle );
}


void
OutALSAStream::init( Frame & f ) throw(exception)
{
    TRACE( "Initializing ALSA outputplugin" );

    if ( getState() == state_error ) return;

    int format = AFMT_S16_NE;
    int retval = -1;
    unsigned int frequency_request = 0;
    unsigned int channels_request = 0;

    InAVStreamIFace * aContext = f.getAudioContext();
    if ( !aContext ) {
        aContext = f.getAVContext();
        if ( !aContext ) {
            INFO( "no audio/av context" );
            return;
        }
    }

    /* get audio properties */
    PropertyNode * av_props = aContext->getProperties();
    PropertyNode * audio = 0;
    PropertyNode * p = 0;
    int d = 0;
    string type;
    while( ( p = av_props->getDaughter( "codec", d++, true ) ) && !audio ) {
        if ( p->getAttr( "type", type ) ) {
            if ( type == "audio" ) {
                audio = p;
            }
        }
    }
    if ( !audio ) {
        WARN( "Couldn't get properties of audio codec" );
        setState( state_error );
        return;
    }
    audio->getAttr( "frequency", retval );
    if ( retval>=0 ) frequency_request = retval;
    audio->getAttr( "channels", retval );
    if ( retval>=0 ) channels_request = retval;
    audio->getAttr( "device", _deviceName );

    TRACE( "requesting: "<<PRINTV( frequency_request )<<PRINTV( channels_request )
           <<" at device: "<<PRINTV(_deviceName) );
    PDELETE( av_props );

    /* sanity checks */
    if ( frequency_request <= 0  ||
         channels_request < 0 )
    {
        WARN( "couldn't get all parameters"
              <<PRINTV(frequency_request)<<PRINTV(channels_request) );
        setState( state_error );
        return;
    }

    /* open ALSA */
    int err;
    int dir;
    snd_pcm_hw_params_t  * hwParams;
    unsigned int frequency_recieved;
    unsigned int channels_recieved;


    if ((err = snd_pcm_open (&_playbackHandle, _deviceName.c_str(), SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
        WARN( "cannot open audio device"
              <<_deviceName<<snd_strerror(err));
        setState( state_error );
        return;
    }
		   
    if ((err = snd_pcm_hw_params_malloc (&hwParams)) < 0) {
        WARN( "cannot allocate hardware parameter structure"
              << snd_strerror (err));
        setState( state_error );
        return;
    }
				 
    if ((err = snd_pcm_hw_params_any (_playbackHandle, hwParams)) < 0) {
        WARN( "cannot initialize hardware parameter structure"
              << snd_strerror (err));
        setState( state_error );
        return;
    }
	
    if ((err = snd_pcm_hw_params_set_access (_playbackHandle, hwParams, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
        WARN( "cannot set access type"
              << snd_strerror (err));
        setState( state_error );
        return;
    }
	
    if ((err = snd_pcm_hw_params_set_format (_playbackHandle, hwParams, SND_PCM_FORMAT_S16_LE)) < 0) {
        WARN( "cannot set sample format"
              << snd_strerror (err));
        setState( state_error );
        return;
    }

	frequency_recieved = frequency_request;
    if ((err = snd_pcm_hw_params_set_rate_near (_playbackHandle, hwParams, &frequency_recieved, &dir)) < 0) {
        WARN( "cannot set sample rate"
              << snd_strerror (err));
        setState( state_error );
        return;
    }
    TRACE( PRINTV( frequency_request ) 
           << PRINTV( frequency_recieved ) );
    if ( frequency_recieved != frequency_request ) {
        WARN( "frequency_recieved != frequency_request "
              << PRINTV( frequency_request ) 
              << PRINTV( frequency_recieved ) );
    }
	_frequency = frequency_recieved;

    if ((err = snd_pcm_hw_params_set_channels (_playbackHandle, hwParams, channels_request)) < 0) {
        WARN( "cannot set channel count"
              << snd_strerror (err));
        setState( state_error );
        return;
    }
    if ((err = snd_pcm_hw_params_get_channels( hwParams, &channels_recieved )) < 0 ){
        WARN( "cannot read channels" );
        setState( state_error );
        return;
    }
    TRACE( PRINTV( channels_request ) 
           << PRINTV( channels_recieved ) );
    if ( channels_recieved != channels_request ) {
        WARN( "channels_recieved != channels_request "
              << PRINTV( channels_request ) 
              << PRINTV( channels_recieved ) );
    }
    _nChannels = channels_recieved;
	
    if ((err = snd_pcm_hw_params (_playbackHandle, hwParams)) < 0) {
        WARN( "cannot set parameters"
              << snd_strerror (err));
        setState( state_error );
        return;
    }
	
    snd_pcm_hw_params_free( hwParams );
	

    if ((err = snd_pcm_prepare (_playbackHandle)) < 0) {
        WARN( "cannot prepare audio interface for use"
              << snd_strerror (err));
        setState( state_error );
        return;
    }

    /* everything went ok */
    setState( state_open );

    return;
}


void 
OutALSAStream::putFrame( Frame & f )
{

    AudioBuffer * a = f.makeABuf();
    if ( !a ) return;

    if ( a->getNChannels() != _nChannels ) {
        if ( _warnCount >0 ) {
            _warnCount--;
            WARN( "#"<<_warnCount<<" audio input nChannels != my nchannels "
                  <<PRINTV( a->getNChannels() ) << PRINTV( _nChannels ) );
        }
        return;
    }

    /*   either the channels are in sepearete buffers, or interleaved
     *   in a single buffer. If they are in seperate buffers, we are
     *   going to interleave them into a single buffer
     *   
     *   FIXME this should be removed since ALSA can handle
     *   non-interleaved audio!  but since we have to expect different
     *   input streams, this requires reconfiguration of the alsa
     *   stream
     */
    int16_t * buffer = (int16_t*)a->interleave()->buf();

    /* Send the audio to the device */
    int total = a->getNSamples();
    int written = 0;
    while ( written < total) {
        int result = snd_pcm_writei( _playbackHandle, buffer + written, total - written);
        if ( result <= 0 ) {
            if ( _warnCount >0 ) {
                _warnCount--;
                WARN( "#"<<_warnCount<<" writing to "<<_deviceName<<PRINTV( result ) );
            }
            return;
        }
        written += result;
    }
    
    return;

}

PropertyNode * 
OutALSAStream::getProperties( PropertyNode * p ) const
{
    WARN( "FIXME" );
    return 0;
}
