/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, Rolf.Dubitzky@gmx.net
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

// STDC
#include <unistd.h>
#include <getopt.h>

// STDC++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// libxml
#include <libxml/parser.h>

// PIAVE
#include "libpiave/piave_base.hh"
#include "libpiave/player.hh"
#include "libpiave/mediaelements.hh"
#include "libpiave/exception.hh"
#include "libpiave/property.hh"
#include "libpiave/plugin.hh"
#include "libpiave/pluginmgr.hh"
#include "libpiave/streammgr.hh"
#include "libpiave/bufferpool.hh"
#include "libpiave/VEML.hh"
#include "libpiave/scenebuilder.hh"

using namespace PIAVE;

StoryBoard * openVEMLFile( string filename );
StoryBoard * openVideoFile( string filename );

void usage( int argc,  char * const * argv  ) { 
    cout << "usage:  "<<argv[0]<<" [options] <file>"<<endl;
    cout << " -v <level>       level of verbosity (default: 0=quiet) " << endl;
    cout << " -d <dir>         search <dir> recursively for plugins" << endl;
    cout << " -l <lib>         load plugin library <lib>" << endl;
    cout << " -V <format>      screen none/rgb/xv" << endl;
    cout << " -A <format>      audio output auto/none/oss/arts" << endl;
    cout << " -f <frames>      total number of frames to show" << endl;
    cout << " -s <scale>       scale with image with factor, e.g. half size: 0.5" << endl;
    cout << " -S <speed>       speed; (0.0 - inf) default 1.0" << endl;
    exit( 1 ); 
}

int
main( int argc,  char * const * argv  )
{

    string video_formatname     = "xv";
    string audio_formatname     = "auto";
    bool   default_plugin_src   = true;
    int    n_frames             = -1;
    int    n_plugins            = 0;
    double scale                = 1.0;
    double speed                = 1.0;

    int c;
    opterr = 0;
    while ((c = getopt (argc, argv, "d:l:f:A:V:s:S:v:qh")) != -1) {
        switch (c)
        {
          case 'v':
              PIAVE::Global::verbosity = atoi( optarg );
              break;
          case 'h':
              usage( argc, argv );
              break;
          case 'd':
              n_plugins += PluginMgr::openDLLsFromDir( optarg );
              default_plugin_src = false;
              break;
          case 'l':
              n_plugins += PluginMgr::openDLL( optarg ); 
              default_plugin_src = false;
              break;
          case 'V':
              video_formatname = string( optarg );
              break;
          case 'A':
              audio_formatname = string( optarg );
              break;
          case 'f':
              n_frames = atoi( optarg );
              break;
          case 's':
              scale = atof( optarg );
              break;
          case 'S':
              speed = atof( optarg );
              break;
          default:
              cout <<"unknown option: "<<(char)c<<endl;
              usage( argc, argv );
              break;
        }
    }

    if ( optind == argc ) {
        cout << "Missing filename" << endl;
        usage( argc, argv );
    }

    string filename( argv[optind] );

    /*
     *    init PIAVE 
     */
    PIAVE::Init();
    PluginMgr::openDLLsFromDefaultPath();


    /*
     *    try to open <file> as a VEML or Video file
     */
    StoryBoard * story = openVEMLFile( filename );
    if ( !story ) {
        story = openVideoFile( filename );
    }
    if ( !story ) {
        cout << "ERROR : couldn't open file \""<<filename<<"\"" << endl;
        return 1;
    }
    
//     Global::display_width  = (int)( Global::render_width  * scale );
//     Global::display_height = (int)( Global::render_width  / Global::render_aspect * scale );

    /*
     *   try to get screen output plugin
     */
    INFO( "looking for video output plugin "<<PRINTV(video_formatname) );
    OutAVStreamIFace * voutstream = 0;
    if ( video_formatname == "none" ) {
    } else {
        PropertyNode * screen_caps = new PropertyNode( "outstream" );
        PropertyNode * monitor = screen_caps->addDaughter( "monitor" );
        PropertyNode * video = monitor->addDaughter( "video" );
        video->addAttr( "type", video_formatname );
        INFO( PRINTV( video_formatname ) );
        Plugin * p = PluginMgr::find( screen_caps );
        if ( !p ) {
            cout << "WARNING: no output plugin to handle output to screen" << endl;
        } else {
            video->addAttr( "width" , Global::renderFmt.width  * scale );
            video->addAttr( "height" , Global::renderFmt.width  / Global::renderFmt.aspect * scale );
            PluginBase * b = p->newElement( video );
            if ( !b ) {
                cout << "WARNING : instantiating plugin "<<p->getName()<<endl;
            } else {
                //  FIXME why doesn't this work ?? 
//                    voutstream = dynamic_cast<OutAVStreamIFace*>( b );
                voutstream = static_cast<OutAVStreamIFace*>( b );
                if ( voutstream==0 ) {
                    cout << "WARNING : wrong plugin type " << endl;
                }
            }
        }
        PDELETE( screen_caps );
    }

    /*
     *   try to get audio output plugin
     */
    INFO( "looking for audio output plugin "<<PRINTV(audio_formatname) );
    OutAudioStreamIFace * aoutstream = 0;
    if ( audio_formatname == "none" ) {
    } else {
        /* define capabilities to look for */
        PropertyNode * speaker_caps = new PropertyNode( "outstream" );
        PropertyNode * monitor = speaker_caps->addDaughter( "monitor" );
        PropertyNode * audio = monitor->addDaughter( "audio" );
        if ( audio_formatname != "auto" ) {
            audio->addAttr( "type", audio_formatname );
        }
        /* try to get plugin */
        bool keep_trying = true;
        while ( keep_trying ) {
            Plugin * p = 0;
            try {
                p = PluginMgr::find( speaker_caps );
                if ( !p ) {
                    cout << "WARNING: no output plugin to handle audio" << endl;
                    keep_trying = false;
                } else {
                    PluginBase * b = p->newElement();
                    //  FIXME why doesn't this work ?? 
//                    voutstream = dynamic_cast<OutAVStreamIFace*>( b );
//                    if ( aoutstream==0 ) {
//                        cout << "WARNING : wrong plugin type " << endl;
//                    }
                    aoutstream = static_cast<OutAudioStreamIFace*>( b );
                    keep_trying = false;
                }
            } catch ( const stream_error & e ) {
                if ( p ) {
                    INFO( "instantiation of plugin \""
                          <<p->getName()<<"\" failed -> disabling" );
                    p->setEnabled( false );
                }
            }  catch ( const std::exception & e ) {
                cerr << "caught std exception: "<<e.what() << endl;
                return 0;
            }
        }
        PDELETE( speaker_caps );
    }


    /*
     *    setup player
     */
    Player player;
    player.setStoryBoard( story );
    player.setVideoOut( voutstream );
    player.setAudioOut( aoutstream );
//    player.setStop( n_frames-1 );
    player.setSpeed( speed );


    /*
     *    main loop
     */
    BufferPool::printStats();
    INFO( "Starting main player loop" );
    player.switchToPlay();
    while ( n_frames!=0 && player.getState() == Player::playing ) {
//        INFO( PRINTV( n_frames ));
        BufferPool::printStats();
        player.playNextFrame();
        n_frames--;
    }


    /*
     *    print some stats
     */
    BufferPool::printStats();
    story->printStats();


    /* 
     *    clean up 
     */
    PDELETE( story );

    return 0;
}


StoryBoard *
openVEMLFile( string filename )
{
    /*
     *  all functionality here must be moved to more proper locations
     */
    INFO( "trying to open VEML file" );

    xmlDocPtr cmndDoc = xmlParseFile( filename.c_str() );
    if ( !cmndDoc ) {
        INFO( "couldn't parse command ... no XML ?" );
        return 0;
    }

    xmlNodePtr messageRoot = xmlDocGetRootElement( cmndDoc );
    if ( !messageRoot ) {
        xmlFreeDoc( cmndDoc );
        INFO( "no root node ... no XML?" );
        return 0;
    }

    StorySeqNode * scenes = 0;
    for ( xmlNodePtr xmlCmndNode = messageRoot->xmlChildrenNode;
          xmlCmndNode != 0;
          xmlCmndNode = xmlCmndNode->next )
    {
        if ( !xmlStrcmp(xmlCmndNode->name, (const xmlChar *)"scenelist") ) {
            VEMLCmnd * cmnd = new VEMLCmnd( "scenelist" );
            cmnd->copyFromXMLNode( cmndDoc, xmlCmndNode );
	    INFO( cmnd->toXML() );
            scenes = SceneBuilder::buildScenelist( cmnd );
            delete cmnd;
        } else if ( !xmlStrcmp(xmlCmndNode->name, (const xmlChar *)"outstream") ) {
            /* parse outstream properties */
            WARN( "FIXME" );
        } else if ( !xmlStrcmp(xmlCmndNode->name, (const xmlChar *)"head") ) {
            /* FIXME move t a better place */
            /*
             *   <head>
             *     <path>/some/absolute/path/to/files</path>
             *     <path>/some/other/absolute/path/to/files</path>
             *   </head>
             *   StreamMgr::addPathEntry( );
             */
        }
    }
    
    if ( !scenes ) return 0;

    xmlFreeDoc( cmndDoc );
 
    /*
     *    setup render tree
     */
    StoryBoard * story = new StoryBoard( "MasterBoard" );
    story->setHead( scenes );
//    story->setProperties( props );

    return story;
}


StoryBoard * 
openVideoFile( string filename )
{
    InAVFStream * instream = 0;
    try {
        instream = StreamMgr::getAVInput( filename );
    } catch ( const file_error & e ) {
        cerr << "caught PIAVE file error: "<<e.what() << endl;
        return 0;
    } catch ( const std::exception & e ) {
        cerr << "caught std exception: "<<e.what() << endl;
        return 0;
    }
    
    if ( !instream ) {
        cerr << " Couldn't open file "<<filename<<endl;
        return 0;
    }

    /*
     *    setup render tree
     */
    AVChunk * avChunk = new AVChunk( instream );
    StoryBoard * story = new StoryBoard( "MasterBoard" );
    story->setHead( avChunk );
//    story->setProperties( props );

    return story;
}
